%-- This code has been realized to illustrate neural network concepts during 
%-- the workshop organized by the "club EEA" which happened in March 2018 
%-- https://rnn2018lyon.sciencesconf.org/
%--
%-- This code has been strongly inspired by the Mooc of Andrew Ng on coursera
%-- https://www.coursera.org/learn/machine-learning/
%-- 
%-- The application treated here concerns the automatic recognition of 
%-- digits drawn by hand. This code is not protected by any license, you 
%-- can re-use it freely without any restriction. Please do not forget to
%-- cite this event for any use of the code.

function [J,grad] = costFunction(nn_params, ...
                                   input_layer_size, ...
                                   hidden_layer_size, ...
                                   output_layer_size, ...
                                   X, y, lambda)
                               
    %-- COSTFUNCTION Implements the neural network cost function 
    %-- for a two layer neural network which performs classification
    %--   [J,grad] = costFunction(nn_params, input_layer_size, ...
    %--   hidden_layer_size, output_layer_size, X, y, lambda) computes the cost
    %--   and gradient of the neural network. The parameters for the neural 
    %--   network are "unrolled" into the vector nn_params and need to be 
    %--   converted back into the weight matrices.
    %-- 
    %--   The returned parameter grad should be a "unrolled" vector of the
    %--   partial derivatives of the neural network.
    

    %-- Reshape nn_params back into the parameters Theta1 and Theta2, the weight matrices
    %-- for our 2 layer neural network
    Theta1 = reshape(nn_params(1:hidden_layer_size * (input_layer_size + 1)), ...
                     hidden_layer_size, (input_layer_size + 1));

    Theta2 = reshape(nn_params((1 + (hidden_layer_size * (input_layer_size + 1))):end), ...
                     output_layer_size, (hidden_layer_size + 1));

    %-- Setup some useful variables
    m = size(X, 1);

    %-- Initialization of energy value J
    J = 0;
    
    %-- Initialization the gradients Theta1_grad and Theta2_grad
    Theta1_grad = zeros(size(Theta1));
    Theta2_grad = zeros(size(Theta2));

    
    %------------------------------------------------------------------
    %------------------------------------------------------------------
    %------------------------------------------------------------------        
    
    %-- Feedforward the neural network and return the cost function 
    %-- without any regularization
    a1 = [ones(m,1) X];
    a2 = nn.sigmoid(Theta1*a1')';
    a2 = [ones(m,1) a2];
    a3 = nn.sigmoid(Theta2*a2')';
    for i=1:m
        yi = zeros(1,size(a3,2));
        yi(y(i)) = 1;
        val = 0;
        for k=1:size(yi,2)
            val = val + yi(k)*log(a3(i,k)) + (1-yi(k))*log(1-a3(i,k));
        end
        J = J + val;
    end
    J = -J/m;

    %-- Cost function with regularization term
    J = J + (lambda/2/m) * ( sum(sum(Theta1(:,2:end).^2)) + ...
        sum(sum(Theta2(:,2:end).^2)) ); 


    %------------------------------------------------------------------
    %------------------------------------------------------------------
    %------------------------------------------------------------------    
        
    %-- Implement the backpropagation algorithm to compute the gradients
    %-- Theta1_grad and Theta2_grad. 
    %-- Return the partial derivatives of the cost function with respect 
    %-- to Theta1 and Theta2 in Theta1_grad and Theta2_grad, respectively.
    for i=1:m

        %-- step 1
        a1 = [1 X(i,:)]';
        z2 = (Theta1*a1);               
        a2 = [1;nn.sigmoid(z2)];
        z3 = (Theta2*a2);
        a3 = nn.sigmoid(z3);
        
        %-- step 2
        Y = zeros(size(a3));
        Y(y(i)) = 1;
        delta3 = a3 - Y;

        %-- step 3        
        delta2 = ((Theta2')*delta3);
        delta2 = delta2(2:end);
        delta2 = delta2.*nn.sigmoidGradient(z2);
        

        %-- step 4
        Theta1_grad = Theta1_grad + delta2*(a1');        
        Theta2_grad = Theta2_grad + delta3*(a2');


    end

    %-- step 5
    Theta1_grad = Theta1_grad / m;
    Theta2_grad = Theta2_grad / m;
    
    
    %-- Add regularization to gradient
    Theta1_grad(:,2:end) = Theta1_grad(:,2:end) + lambda/m * Theta1(:,2:end);
    Theta2_grad(:,2:end) = Theta2_grad(:,2:end) + lambda/m * Theta2(:,2:end);


    %-- Unroll gradients
    grad = [Theta1_grad(:) ; Theta2_grad(:)];

end

